/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * PixlScan.H
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 02/11/90
 *   MODIFIED : 07/26/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Macros and externs that allow flipping and color remapping
 *		of an unpacked raw mcga shape as it is scanned from a buffer.
 *
 * HISTORY
 *
*/
#ifndef ECHIDNA_PIXLSCAN_H
#define ECHIDNA_PIXLSCAN_H

#include <echidna/platform.h>
#include "switches.h"

#ifdef __cplusplus
extern "C" {
#endif

#include <echidna/etypes.h>
#include <echidna/graphics.h>

/**************************** C O N S T A N T S ***************************/
#define NUM_COLORS	255

/******************************** T Y P E S *******************************/

/****************************** G L O B A L S *****************************/

/******************************* M A C R O S ******************************/

/*********************************************************************
 *
 * ResetPixlScan
 *
 * SYNOPSIS
 *		#define ResetPixlScan(inshape, Hflip, Vflip, outshape)							\
 *
 * PURPOSE
 *		To Set up the global variables of PixlScan to scan the shape in 
 *		a direction that will result in output flipped to the specifications
 *		in the 'attrib' flags.
 *
 * INPUT
 *		inshape:	(GShape *)	Unpacked Shape to translate.
 *		Hflip:		(short)		True if brush should be flipped horizontally.
 *		Vflip:		(short)		True if brush should be flipped vertically.
 *		outshape:	(UBYTE	*)	Work Buffer for doing translation.
 *	
 *
 * EFFECTS
 *		XOffset and YOffset of 'outshape' are adjusted to take flipping into
 *		account.
 *		Sets Values Of Globals:
 *			ps_Size			<- Area of trimmed sprite.
 *			ps_ColCount		<- 0
 *			ps_DWidth		<- trimmed width of 'inshape'
 *			ps_2xDWidth		<- 2 * ps_DWidth
 *			ps_PixlPtr		<- address of pixel at which to start scanning.
 *			ps_DeltaCol		<- direction to scan a row (+1 or -1).
 *			ps_DeltaRow		<- value to change ps_PixlPtr by at end of row to
 *								get to start pixel of next row.
 *
 * RETURN VALUE
 *		Undefined
 *
 * HISTORY
 *	
 *
 * SEE ALSO
 *
*/
#define ResetPixlScan(inshape, isHflipped, isVflipped)						\
{																			\
	ps_DWidth = (inshape)->Data.MCGA.DWidth;								\
	ps_2xDWidth = 2 * ps_DWidth;											\
																			\
	ps_Size = (inshape)->Size;												\
	ps_ColCount = 0;														\
	ps_PixlPtr = (UBYTE *)(&((inshape)->Data)) + MCGASDATASIZE;				\
	ps_DeltaCol = 1;														\
	ps_DeltaRow = 0;														\
	if (Hflip)				 												\
	{			 															\
		ps_DeltaCol = -1;													\
		ps_DeltaRow = ps_2xDWidth;											\
		ps_PixlPtr += (ps_DWidth - 1);										\
	}																		\
	if (Vflip)				 												\
	{			 															\
		ps_DeltaRow = -(ps_DeltaRow ^ ps_2xDWidth);							\
		ps_PixlPtr += ps_DWidth * ((inshape)->Data.MCGA.DHeight - 1);		\
	}																		\
} /* ResetPixlScan */ 

/*********************************************************************
 *
 * SetColorMap
 *
 * SYNOPSIS
 *		#define SetColorMap(remap, ColorMap)
 *
 * PURPOSE
 *		To Set up the Color remapping table for scanning and remapping.
 *
 * INPUT
 *		remap:	(RemapTableType *) 
 *		ColorMap:	(UBYTE	*)
 *
 * EFFECTS
 *		Changes values of array ColorMap.
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		ClearColorMap()
*/
#define SetColorMap(remap, ColorMap)										\
{																			\
	short	r;																\
	UBYTE	*from	= (remap)->From;										\
	UBYTE	*to		= (remap)->To;											\
																			\
	for (r = 0; r < (remap)->Elements; r++) {								\
		(ColorMap)[*from++] = *to++;										\
	}																		\
} /* SetColorMap */

/*********************************************************************
 *
 * ClearColorMap
 *
 * SYNOPSIS
 *		#define ClearColorMap(remap, ColorMap)
 *
 * PURPOSE
 *		To restore values changed in previous call to SetColorMap with the
 *		parameter 'remap'.
 *
 * INPUT
 *		remap:	(RemapTableType *) 
 *		ColorMap:	(UBYTE	*)
 *
 * EFFECTS
 *		Changes values of array ColorMap.
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		SetColorMap()
*/
#define ClearColorMap(remap, ColorMap)										\
{																			\
	short	r;																\
	UBYTE	*from	= (remap)->From;										\
																			\
	for (r = (remap)->Elements; r ; r--) {									\
		(ColorMap)[*from] = *from++;										\
	}																		\
} /* ClearColorMap */
	
/*********************************************************************
 *
 * ResetPatternMap
 *
 * SYNOPSIS
 *		#define ResetPatternMap(patmap)
 *
 * PURPOSE
 *		To Set up the pattern remapping for pattern remapping during scan.
 *
 * INPUT
 *		patmap:	(PatmapTableType *)
 *
 * EFFECTS
 *		Changes values of global array ps_PatternMap.
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
#define ResetPatternMap(patmap)				\
{											\
} /* ResetPatternMap */

/*********************************************************************
 *
 * ClosePixlScan
 *
 * SYNOPSIS
 *		#define ClosePixlScan(cs)
 *
 * PURPOSE
 *		To Free the tables allocated by OpenPixlScan().
 *
 * INPUT
 *		cs:	(ConvertSType *)
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
#define ClosePixlScan(ptr,size)							\
{														\
	DestroyTempBuffer((ptr)->ColorMap, size);			\
} /* ClosePixlScan */


/*********************************************************************
 *
 * GetPixl
 *
 * SYNOPSIS
 *		#define GetPixl()
 *
 * PURPOSE
 *		To return the next pixel scanned from the source shape.
 *
 * INPUT
 *		NONE.
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		If has not reached end of shape then returns positive remapped 
 *			pixel value,
 *		Else returns -1.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
#define GetPixl()											\
{															\
	--ps_Size;												\
	ps_Pixl = ps_ColorMap[*ps_PixlPtr];						\
	if (++ps_ColCount < ps_DWidth) {						\
		ps_PixlPtr += ps_DeltaCol;							\
	} else {												\
		ps_ColCount = 0;									\
		ps_PixlPtr += ps_DeltaCol + ps_DeltaRow;			\
	}														\
} /* GetPixl */

/****************** F U N C T I O N   P R O T O T Y P E S *****************/
#if __NOT_ANSI__		/** Non-ANSI Prototypes **/


#else					/** ANSI Prototypes **/


#endif /** Function Prototypes **/
extern short OpenPixlScan (
	ConvertSType	*cs,
	short	maxcolors
);																		

#ifdef __cplusplus
}
#endif

#endif /* ECHIDNA_PIXLSCAN_H */

