/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * TREES.C
 *
 * PROGRAMMER : R.G. Marquez
 *    VERSION : 01.000
 *    CREATED : 08/25/89
 *   MODIFIED : 07/27/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *          Library Module, Tree support.
 *
 * HISTORY
 *	08/25/89 Friday - Created. (RGM)
 *	01/26/90 Friday - Added required prototypes (...FuncPtr). (GAT)
 *
*/

#include <echidna/platform.h>

#include <stdio.h>
#include <stdlib.h>
#include <echidna/etypes.h>
#include <echidna/sysextra.h>

#include <echidna/trees.h>

#if __NOT_ANSI__
extern void *calloc ();
extern void free ();
#else
#endif
/**************************** C O N S T A N T S ***************************/


/******************************** T Y P E S *******************************/


/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/


/*------------------------------------------------------------------------*/
/**# MODULE:TREES_BSTfind                                                 */
/*------------------------------------------------------------------------*/

void *BSTfind (A, x, comparefn)

	TreeNodePtrType *A;
	void *x;
	TreeCFuncPtr comparefn;
{
	void *element;
	void *result = NULL;

	element = TBSTfind (A, x, comparefn);
	if (element) {
		result = DATA (element);
	}
	return (result);
}

/*------------------------------------------------------------------------*/
/**# MODULE:TREES_BSTinsert                                               */
/*------------------------------------------------------------------------*/

BOOL BSTinsert (A, x, comparefn)

	TreeNodePtrType *A;
	void *x;
	TreeCFuncPtr comparefn;
{
	BOOL status = FALSE;
	int sign = 1;


	while (*A) {

		sign = (*comparefn) (x, DATA(*A));

		if (sign < 0) {
			A = &(LEFTCHILD(*A));
		}

		else {
			A = &(RIGHTCHILD(*A));
		}

	}
	
	NewTreeNode (A, x);

	if (*A) {					/* Make sure we actually have one */
		status = TRUE;
	}
	
	return (status);
}

/*------------------------------------------------------------------------*/
/**# MODULE:TREES_BSTremove                                               */
/*------------------------------------------------------------------------*/

void *BSTremove (A, x, comparefn)

	TreeNodePtrType *A;
	void *x;
	TreeCFuncPtr comparefn;
{
	void *element;
	void *result = NULL;

	element = TBSTremove (A, x, comparefn);
	if (element) {
		result = DATA (element);
	}
	return (result);
}

/*------------------------------------------------------------------------*/
/**# MODULE:TREES_BTclose                                                 */
/*------------------------------------------------------------------------*/

void BTclose (A, comparefn, disposefn)

	TreeNodePtrType *A;
	TreeCFuncPtr comparefn;
	TreeDFuncPtr disposefn;
{
	if (*A) {

		/*** first deallocate the left and right nodes ***/

		BTclose (&(LEFTCHILD(*A)), comparefn, disposefn);
		BTclose (&(RIGHTCHILD(*A)), comparefn, disposefn);

		/*** now deallocate this node's memory space ***/

		(*disposefn) (DATA(*A));
		BTfree(A);
	}

	return;
}

/*------------------------------------------------------------------------*/
/**# MODULE:TREES_BTprint                                                 */
/*------------------------------------------------------------------------*/

void BTprint (A, printfn)

	TreeNodePtrType *A;
	TreePFuncPtr printfn;
{
	if (*A) {
		if (LEFTCHILD(*A)) {
			BTprint (&(LEFTCHILD(*A)), printfn);
		}

		(*printfn)(DATA(*A));

		if (RIGHTCHILD(*A)) {
			BTprint (&(RIGHTCHILD(*A)), printfn);
		}
	}

	return;
}

/*------------------------------------------------------------------------*/
/**# MODULE:TREES_BTRprint                                                */
/*------------------------------------------------------------------------*/

/*
 * BTRprint.c
 *
 * PROGRAMMER : R.G. Marquez & Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 09/27/89
 *   MODIFIED : 
 *
 * DESCRIPTION
 *          Library Module, Tree support with extra recursion.
 *
 * HISTORY
 *
 *
*/
void BTRprint (A, printfn)

	TreeNodePtrType *A;
	TreeP2FuncPtr printfn;
{
	if (*A) {
		if (LEFTCHILD(*A)) {
			BTRprint (&(LEFTCHILD(*A)), printfn);
		}

		(*printfn)(DATA(*A), (TreePFuncPtr)BTRprint);

		if (RIGHTCHILD(*A)) {
			BTRprint (&(RIGHTCHILD(*A)), printfn);
		}
	}

	return;
}

/*------------------------------------------------------------------------*/
/**# MODULE:TREES_TBSTcopy                                                */
/*------------------------------------------------------------------------*/

/*########################## STATIC VARIABLES ############################*/
static TreeNodePtrType NewTree;
static TreeCPFuncPtr CopyFN;
static TreeCFuncPtr CompareFN;



/*###########################  FUNCTIONS  ################################*/

#if __NOT_ANSI__
extern BOOL rCopyBST ();
#else
extern BOOL rCopyBST (TreeNodePtrType *);
#endif

BOOL rCopyBST (A)

	TreeNodePtrType *A;
{
	void *new_data = NULL;
	BOOL ok = TRUE;

	if (*A) {

		new_data = (*CopyFN) (DATA(*A));

		if (new_data) {

			ok = BSTinsert (&NewTree, new_data, CompareFN);

			if ((LEFTCHILD(*A)) && (ok)){
				ok = rCopyBST (&(LEFTCHILD(*A)));
			}

			if ((RIGHTCHILD(*A)) && (ok)) {
				ok = rCopyBST (&(RIGHTCHILD(*A)));
			}
		}
		else {
			ok = FALSE;
		}
	}

	return (ok);
}


void *TBSTcopy (A, copyfn, comparefn, disposefn)


	TreeNodePtrType *A;
	TreeCPFuncPtr copyfn;
	TreeCFuncPtr comparefn;
	TreeDFuncPtr disposefn;
{
	BOOL ok;


	NewTree = NULL;				/* Initialize static varaibles */
	CompareFN = comparefn;			/* (statics used to speed up   */
	CopyFN = copyfn;				/*  recusion).                 */

	ok = rCopyBST (A);				/* Set off the recursive copy  */



	if (! ok) {
		BSTclose (&NewTree, comparefn, disposefn);
	}

	return (NewTree);
}

/*------------------------------------------------------------------------*/
/**# MODULE:TREES_TBSTfind                                                */
/*------------------------------------------------------------------------*/

void *TBSTfind (A, x, comparefn)

	TreeNodePtrType *A;
	void *x;
	TreeCFuncPtr comparefn;
{
	void *element = NULL;
	int sign = 1;

	while ((*A) && sign) {
		sign = (*comparefn) (x, DATA(*A));

		if (sign < 0) {
			A = &(LEFTCHILD(*A));
		}
		else if (sign > 0) {
			A = &(RIGHTCHILD(*A));
		}
	}
	
	if (! sign) {				/* No difference between node & target */
		element = (*A);
	}
	
	return (element);
}

/*------------------------------------------------------------------------*/
/**# MODULE:TREES_TBSTremove                                              */
/*------------------------------------------------------------------------*/

void *TBSTremove (A, x, comparefn)

	TreeNodePtrType *A;
	void *x;
	TreeCFuncPtr comparefn;
{
	TreeNodePtrType	*topsplice		= NULL;
	TreeNodePtrType	deadnode			= NULL;
	TreeNodePtrType	replacementnode	= NULL;
	void				*element			= NULL;
	int				sign				= 1;


	/* Search for the target node */

	while ((*A) && sign) {
		sign = (*comparefn) (x, DATA(*A));

		if (sign < 0) {
			A = &(LEFTCHILD(*A));
		}
		else if (sign > 0) {
			A = &(RIGHTCHILD(*A));
		}
	}

	if (! sign) {				/* No difference between node & target */

		element = (*A);

		deadnode = *A;

		if ( (LEFTCHILD(deadnode)) && (RIGHTCHILD(deadnode)) ) {
			topsplice = A;				/* Save splice address */
			A = &(RIGHTCHILD(*A));
			while (LEFTCHILD(*A)) {
				A = &(LEFTCHILD(*A));	/* Find least of right */
			}
			replacementnode = *A;		/* Replacement's address */
			*A = RIGHTCHILD(replacementnode);
			*topsplice = replacementnode; /* ..into top            */
			LEFTCHILD(replacementnode)  = LEFTCHILD(deadnode);
			RIGHTCHILD(replacementnode) = RIGHTCHILD(deadnode);
		}

		else if (LEFTCHILD(deadnode)) {		/* Only a left child  */
			*A = LEFTCHILD(deadnode);
		}

		else if (RIGHTCHILD(deadnode)) {		/* Only a right child */
			*A = RIGHTCHILD(deadnode);
		}

		else {						/* No children        */
			*A = NULL;
		}

		BTfree(&deadnode);

	}


	return (element);
}

